/*
** Copyright (c) 1995,1996
**      Michael Smith and Genesis Software.  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer as
**    the first lines of this file unmodified.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY Michael Smith AND Genesis Software ``AS IS'' 
** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED. IN NO EVENT SHALL Michael Smith AND OR Genesis Software BE
** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
** POSSIBILITY OF SUCH DAMAGE.
**
**      $Id: mdsio.c,v 1.11 1997/03/20 04:18:27 msmith Exp $
*/
/*
** RS-485 driver for 8250-family UARTs.
**
** Assorted wisdom, comments and code fragments taken from the FreeBSD 'sio' 
** driver; thanks and credit to all who have worked on it.
**
** See mdsio(4) for a description of the flags values and the programming
** interface.
*/

#include "mdsio.h"
#if NMDSIO > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/malloc.h>
#include <sys/select.h>
#include <sys/fcntl.h>
#include <sys/vnode.h>

#include <machine/mdsio.h>
#include <machine/clock.h>


#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#include <i386/isa/isa.h>
#include <i386/isa/isa_device.h>
#include <i386/isa/sioreg.h>
#include <i386/isa/ic/ns16550.h>

#include "opt_mdsio.h"

/*******************************************************************************
** Determine whether we're a FreeBSD 2.1 strain.  Default is 2.2.
**
** Do this so that if this is really the wrong way, there's only one place to change.
**
** Note that for 2.1 strains, the driver must also be added to conf.c.
*/
#if BSD < 199506
#define FreeBSD_21
#define __unused
#endif

#ifdef DEVFS
#include <sys/devfsext.h>
#endif

/*******************************************************************************
** Handle parameter overrides from the kernel config file
*/
#ifndef MDSIO_RXBUF
#define MDSIO_RXBUF	1024
#else
#if MDSIO_RXBUF > 65536
#error MDSIO_RXBUF too large
#endif
#endif

#ifndef MDSIO_TXBUF
#define MDSIO_TXBUF	1024
#else
#if MDSIO_TXBUF > 65536
#error MDSIO_TXBUF too large
#endif
#endif

#ifndef MDSIO_TIMEOUT
#define MDSIO_TIMEOUT	64
#endif

#ifndef MDSIO_FTL
#define MDSIO_FTL	FIFO_RX_MEDH
#endif

/*******************************************************************************
** Per-port structure and related definitions
*/
typedef struct 
{
    u_short		state;			/* current port state */
#define MD_OPEN		(1<<0)			/* port is open */
#define MD_READSLEEP	(1<<1)			/* sleeping for read */
#define MD_XMITTING	(1<<2)			/* transmitting */
#define MD_PREADDR	(1<<3)			/* waiting for SR to drain before sending address */
#define MD_POSTADDR	(1<<4)			/* waiting for address to be sent before going to data */
#define MD_WRITESLEEP	(1<<5)
#define MD_TIMER	(1<<6)			/* there's a timeout routine active on this interface */
#define MD_TAILDRAIN	(1<<7)			/* waiting for the last byte to drain before disabling Tx */
#define MD_DRAINSLEEP	(1<<8)			/* sleeping for drain so we can close */

    struct selinfo	rsel,wsel;		/* Information for select() */

    u_short		iobase;			/* base port address */
    u_char		hasfifo;
    int			unit;			/* unit number */
    int			eflag;
#define MD_FIFOOF	(1<<0)			/* FIFO overflow */
#define MD_IBUFOF	(1<<1)			/* ringbuffer overflow */
#define MD_FERROR	(1<<2)			/* framing error (noise) */
#define MD_RXWHILETX	(1<<3)			/* received data while transmitting */

    int			speed;
    u_short		options;		/* port configuration options */
#define MD_TXCTL	(1<<0)			/* Tx/Rx control required */
#define MD_2WIRE	(1<<1)			/* 2-wire mode */

    u_char		tx_mcr;			/* MCR values for each state */
    u_char		rx_mcr;
    u_char		cur_mcr;		/* current MCR value */

    u_short		rxbuf[MDSIO_RXBUF];	/* receiver ringbuffer */
    u_short		rxtail,rxhead;

    u_short		txbuf[MDSIO_TXBUF];	/* transmitter ringbuffer */
    u_short		txtail,txhead;

    u_long		ints;			/* total interrupts */
    u_long		rxbytes;		/* bytes received */
    u_long		txbytes;		/* bytes sent */
    u_long		errors;			/* total errors */

#ifdef DEVFS
    void		*devfs_token;
#endif
} MDIF;

MDIF 				mdsio_sc[NMDSIO];

#define mdsio_unit(x)		(mdsio_sc+x)		/* XXX requires contiguous numbers! */

/*******************************************************************************
** Flag decoding macros
*/
#define RTS_TXCONTROL(dev)	((dev)->id_flags & 0x04)
#define DTR_TXCONTROL(dev)	((dev)->id_flags & 0x08)
#define TXCTRL_INVERT(dev)	((dev)->id_flags & 0x10)
#define MDSIO_TXCONTROL(dev)	((dev)->id_flags & 0x0c)
#define MDSIO_2WIRE(dev)	((dev)->id_flags & 0x20)
#define PROBE_VERBOSE(dev)	((dev)->id_flags & 0x80)

/*******************************************************************************
** Buffer manipulation macros
*/
#define rxbuflen(x)		((x->rxhead + MDSIO_RXBUF - x->rxtail) % MDSIO_RXBUF)
#define rxbuffree(x)		(MDSIO_RXBUF - rxbuflen(x))
#define rxheadnext(x)		((x->rxhead+1)%MDSIO_RXBUF)
#define txbuflen(x)		((x->txhead + MDSIO_TXBUF - x->txtail) % MDSIO_TXBUF)
#define txbuffree(x)		(MDSIO_TXBUF - txbuflen(x))
#define txtailnext(x)		((x->txtail+1)%MDSIO_TXBUF)


/*******************************************************************************
** Misc defines/macros
*/
/* Set the mcr for Tx (m==MD_TX) or Rx (m==MD_RX) */
#define set_mcr(x,m)		outb(x->iobase + com_mcr, x->cur_mcr = (m ? x->tx_mcr:x->rx_mcr))
#define MD_TX			1
#define MD_RX			0

/*
** These macros require that all routines have a pointer 'md' pointing
** to the MDIF struct for the current port.
*/
#ifdef MDSIODEBUG
#define debug(fmt, args...)	printf("mdsio%d(%s): " fmt "\n" ,md->unit ,  __FUNCTION__ , ##args)
#else
#define debug(fmt, args...)	
#endif
#define error(fmt, args...)	printf("mdsio%d: " fmt "\n" , md->unit , ##args);

#define MD_ADDRMARK		0x100
#define LCR_NORM		(CFCR_8BITS|CFCR_PENAB|CFCR_PEVEN|CFCR_PONE)	/* normal (listen) mode */
#define LCR_ADDR		(CFCR_8BITS|CFCR_PENAB|CFCR_PODD|CFCR_PONE)	/* addressmark mode */

#define com_scr			7	/* scratch register for 16450-16550 (R/W) */

/*******************************************************************************
** Do we have new-style FTL defines?
*/
#ifndef FIFO_RX_HIGH
#define FIFO_RX_LOW	FIFO_TRIGGER_1
#define FIFO_RX_MEDL	FIFO_TRIGGER_4
#define FIFO_RX_MEDH	FIFO_TRIGGER_8
#define FIFO_RX_HIGH	FIFO_TRIGGER_14
#endif

/*******************************************************************************
** Local prototypes
*/

static int		mdsioattach	__P((struct isa_device *dev));
static int		mdsioprobe	__P((struct isa_device *dev));

static timeout_t	mdsiocomplain;
static timeout_t	mdsiotimeout;

/*******************************************************************************
** ISA administrivia
*/

static char driver_name[] = "mdsio";

struct isa_driver       mdsiodriver = {
    mdsioprobe, mdsioattach, driver_name
};

#ifdef FreeBSD_21
/*******************************************************************************
** Prototype the interrupt handler, devconf cruft
*/

void		mdsiointr	__P((int unit));

#include <sys/devconf.h>

static struct kern_devconf kdc_mdsio[NMDSIO] = 
{ {
        0, 0, 0,                /* filled in by dev_attach */
        driver_name, 0, { MDDT_ISA, 0, "tty" },
        isa_generic_externalize, 0, 0, ISA_EXTERNALLEN,
        &kdc_isa0,              /* parent */
        0,                      /* parentdata */
        DC_UNCONFIGURED,        /* state */
        "Multidrop RS-485 serial port",
        DC_CLS_SERIAL           /* class */
} };

static void
mdsioregister(struct isa_device *dev)
{
    int     unit;

    unit = dev->id_unit;
    if (unit != 0)
	kdc_mdsio[unit] = kdc_mdsio[0];
    kdc_mdsio[unit].kdc_unit = unit;
    kdc_mdsio[unit].kdc_isa = dev;
    dev_attach(&kdc_mdsio[unit]);
}

#else

/*******************************************************************************
** cdevsw stuff, replaces old conf.c mods.
**
*/
static d_open_t		mdsioopen;
static d_close_t	mdsioclose;
static d_read_t		mdsioread;
static d_write_t	mdsiowrite;
static d_ioctl_t	mdsioioctl;
static d_select_t	mdsioselect;

#define CDEV_MAJOR	20
static struct cdevsw mdsio_cdevsw =
{
    mdsioopen,		mdsioclose,	mdsioread,	mdsiowrite,
    mdsioioctl,		nullstop,	nullreset,	nodevtotty,
    mdsioselect,	nommap,		NULL,		driver_name,
    NULL, -1
};
#endif

#ifdef MDSIO_MODULE
/*******************************************************************************
** LKM support
**
*/

#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/lkm.h>

/* presumes standard config of COM3, IRQ 10 */
static struct isa_device mdsio_mod_dev = {
    0, &mdsiodriver, IO_COM3, 10, -1, (caddr_t) 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0
};

MOD_DEV(mdsio, LM_DT_CHAR, CDEV_MAJOR, &mdsio_cdevsw);

static int
mdsio_load(struct lkm_table *lkmtp, int cmd)
{
    if (mdsioprobe(&mdsio_mod_dev)) {
	mdsioattach(&mdsio_mod_dev);
	uprintf("mdsio: driver loaded.\n");
	return 0;
    } else {
	uprintf("mdsio: probe failed.\n");
	return 1;
    }
}

static int
mdsio_unload(struct lkm_table *lkmtp, int cmd)
{
    struct mdsio_softc	*sc;
    int			i;
    
    for (i = 0; i < NMDSIO; i++) {
	sc = mdsio_softc[i];
	if (sc->state & MD_OPEN) {
	    uprintf("mdsio%d: cannot unload (device busy)\n",i);
	    return 1;
	}
	uprintf("mdsio: driver unloaded.\n");
	return 0;
    }
}

static int
mdsio_stat(struct lkm_table *lkmtp, int cmd)
{	
    return 0;
}

int
mdsio_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
#define _module mdsio_module
    DISPATCH(lkmtp, cmd, ver, mdsio_load, mdsio_unload, mdsio_stat);
}
#endif /* MDSIO_MODULE */

/*******************************************************************************
** Device driver functions start here
*/
/*
** mdsioprobe
**
** Renamed and recontextified sioprobe(); find and test the UART.
*/
  
static int
mdsioprobe(struct isa_device *dev)
{
    u_char	mcr_image;
    u_char	failures[10];
    u_short	iobase;
    int		result;
    int		fn;

#ifdef FreeBSD_21
    mdsioregister(dev);			/* administrivia */
#endif    

    mcr_image = MCR_IENABLE;    	/* prepart to set MCR_IENABLE */
    
    bzero(failures, sizeof failures);
    iobase = dev->id_iobase;

    /*
     * Initialize the speed and the word size and wait long enough to
     * drain the maximum of 16 bytes of junk in device output queues.
     * The speed is undefined after a master reset and must be set
     * before relying on anything related to output.  There may be
     * junk after a (very fast) soft reboot and (apparently) after
     * master reset.
     */
    outb(iobase + com_lctl, CFCR_DLAB);
    outb(iobase + com_dlbl, COMBRD(9600) & 0xff);
    outb(iobase + com_dlbh, (u_int) COMBRD(9600) >> 8);
    outb(iobase + com_lctl, CFCR_8BITS);
    DELAY((16 + 1) * 1000000 / (9600 / 10));

    /*
     * Enable the interrupt gate and disable device interupts.  This
     * should leave the device driving the interrupt line low and
     * guarantee an edge trigger if an interrupt can be generated.
     */
    outb(iobase + com_mcr, mcr_image);
    outb(iobase + com_ier, 0);

    /*
     * Attempt to set loopback mode so that we can send a null byte
     * without annoying any external device.
     */
    outb(iobase + com_mcr, mcr_image | MCR_LOOPBACK);
    
    /*
     * Attempt to generate an output interrupt.  On 8250's, setting
     * IER_ETXRDY generates an interrupt independent of the current
     * setting and independent of whether the THR is empty.  On 16450's,
     * setting IER_ETXRDY generates an interrupt independent of the
     * current setting.  On 16550A's, setting IER_ETXRDY only
     * generates an interrupt when IER_ETXRDY is not already set.
     */
    outb(iobase + com_ier, IER_ETXRDY);

    /*
     * On some 16x50 incompatibles, setting IER_ETXRDY doesn't generate
     * an interrupt.  They'd better generate one for actually doing
     * output.  Loopback may be broken on the same incompatibles but
     * it's unlikely to do more than allow the null byte out.
     */
    outb(iobase + com_data, 0);
    DELAY((1 + 2) * 1000000 / (9600 / 10));

    /*
     * Turn off loopback mode so that the interrupt gate works again
     * (MCR_IENABLE was hidden).  This should leave the device driving
     * an interrupt line high.  It doesn't matter if the interrupt
     * line oscillates while we are not looking at it, since interrupts
     * are disabled.
     */
    outb(iobase + com_mcr, mcr_image);

    /*
     * Check that
     *      o the CFCR, IER and MCR in UART hold the values written to them
     *        (the values happen to be all distinct - this is good for
     *        avoiding false positive tests from bus echoes).
     *      o an output interrupt is generated and its vector is correct.
     *      o the interrupt goes away when the IIR in the UART is read.
     */
    failures[0] = inb(iobase + com_lctl) - CFCR_8BITS;
    failures[1] = inb(iobase + com_ier) - IER_ETXRDY;
    failures[2] = inb(iobase + com_mcr) - mcr_image;
    if (dev->id_irq != 0)
	failures[3] = isa_irq_pending(dev) ? 0 : 1;
    failures[4] = (inb(iobase + com_iir) & IIR_IMASK) - IIR_TXRDY;
    if (dev->id_irq != 0)
	failures[5] = isa_irq_pending(dev) ? 1 : 0;
    failures[6] = (inb(iobase + com_iir) & IIR_IMASK) - IIR_NOPEND;

    /*
     * Turn off all device interrupts and check that they go off properly.
     * Leave MCR_IENABLE alone, it gates the OUT2 output of the UART to
     * the ICU input.  Closing the gate would give a floating ICU input
     * (unless there is another device driving it) and spurious interrupts.
     * (On the system that this was first tested on, the input floats high
     * and gives a (masked) interrupt as soon as the gate is closed.)
     */
    outb(iobase + com_ier, 0);
    outb(iobase + com_lctl, CFCR_8BITS);    /* dummy to avoid bus echo */
    failures[7] = inb(iobase + com_ier);
    if (dev->id_irq != 0)
	failures[8] = isa_irq_pending(dev) ? 1 : 0;
    failures[9] = (inb(iobase + com_iir) & IIR_IMASK) - IIR_NOPEND;

    result = IO_COMSIZE;
    for (fn = 0; fn < (int)sizeof(failures); ++fn)
	if (failures[fn]) {
	    outb(iobase + com_mcr, 0);
	    result = 0;
	    if (PROBE_VERBOSE(dev))
		printf("mdsio%d: probe test %d failed\n",
		       dev->id_unit, fn);
	}
    return (result);
}

/*
** mdsioattach
**
** Set ourselves up to talk to the port.  Again derived from the
** corresponding sio routine.
*/

static int
mdsioattach(struct isa_device *isdp)
{
    int		unit;
    MDIF	*md;
    u_char 	scr;
    u_char  	scr1;
    u_char  	scr2;
    u_short	iobase;
    dev_t	dev;
    
    isdp->id_ri_flags |= RI_FAST;
    unit = isdp->id_unit;
    md = mdsio_unit(unit);
    md->unit = unit;				/* save for reverse lookup */
    md->ints = md->errors = md->rxbytes = 0;	/* zero stats */
    
    /*
     * mdsioprobe() has initialized the device registers as follows:
     *      o cfcr = CFCR_8BITS.
     *        It is most important that CFCR_DLAB is off, so that the
     *        data port is not hidden when we enable interrupts.
     *      o ier = 0.
     *        Interrupts are only enabled when the line is open.
     *      o mcr = MCR_IENABLE
     */

    md->rxhead = md->rxtail = 0;	/* initialise the ringbuffer */
    iobase = md->iobase = isdp->id_iobase;
    md->hasfifo = FALSE;		/* at least, so we assume */

    md->options = 0;
    if (MDSIO_2WIRE(isdp))		/* 2-wire mode? */
	md->options |= MD_2WIRE;
    if (MDSIO_TXCONTROL(isdp)) {	/* Tx/Rx control required? */
	u_char	ass = 0;
	debug("tx control nominated");
	
	md->options |= MD_TXCTL;
	md->tx_mcr = md->rx_mcr = 0;
	if (DTR_TXCONTROL(isdp))
	    ass = 0x10;
	if (RTS_TXCONTROL(isdp))	/* RTS more common, assume in donkey case */
	    ass = 0x20;
	if (TXCTRL_INVERT(isdp)) {	/* flipped? */
	    md->tx_mcr = MCR_IENABLE;
	    md->rx_mcr = ass | MCR_IENABLE;
	} else {
	    md->tx_mcr = ass | MCR_IENABLE;
	    md->rx_mcr = MCR_IENABLE;
	}
    } else {
	md->tx_mcr = md->rx_mcr = MCR_IENABLE;
    }
    set_mcr(md,MD_RX);		/* go immediately to receive mode */

    /* attempt to determine UART type */
    printf("mdsio%d: type", unit);
   
    scr = inb(iobase + com_scr);
    outb(iobase + com_scr, 0xa5);
    scr1 = inb(iobase + com_scr);
    outb(iobase + com_scr, 0x5a);
    scr2 = inb(iobase + com_scr);
    outb(iobase + com_scr, scr);
    if (scr1 != 0xa5 || scr2 != 0x5a) {
	printf(" 8250");
#ifdef FreeBSD_21
	kdc_mdsio[unit].kdc_description =
	    "Multidrop interface: National 8250 or compatible";
#endif
    } else {
	outb(iobase + com_fifo, FIFO_ENABLE | FIFO_RX_HIGH);
	DELAY(100);
	switch (inb(iobase + com_iir) & IIR_FIFO_MASK) {
	case FIFO_RX_LOW:
	    printf(" 16450");
#ifdef FreeBSD_21
	    kdc_mdsio[unit].kdc_description =
		"Multidrop interface: 16450 or compatible";
#endif
	    break;
	case FIFO_RX_MEDL:
	    printf(" 16450?");
#ifdef FreeBSD_21
	    kdc_mdsio[unit].kdc_description =
		"Multidrop interface: maybe 16450 (funny FIFO settings)";
#endif
	    break;
	case FIFO_RX_MEDH:
	    printf(" 16550?");
#ifdef FreeBSD_21
	    kdc_mdsio[unit].kdc_description =
		"Multidrop interface: maybe 16550 (funny FIFO settings)";
#endif
	    break;
	case FIFO_RX_HIGH:
	    printf(" 16550A");
	    md->hasfifo = TRUE;
#ifdef FreeBSD_21
	    kdc_mdsio[unit].kdc_description =
		"Multidrop interface: 16550 or compatible";
#endif

	    break;
	}
	outb(iobase + com_fifo, 0);
    }

    if (md->options & MD_TXCTL)
	printf(", %d-wire mode,%s %s transmit control.\n",
	       (md->options & MD_2WIRE) ? 2:4,
	       TXCTRL_INVERT(isdp) ? " inverted":"",
	       DTR_TXCONTROL(isdp) ? "DTR":"RTS");
    else
	printf(", %d-wire mode, automatic transmit control.\n",
	       (md->options & MD_2WIRE) ? 2:4);

#ifdef FreeBSD_21
    kdc_mdsio[unit].kdc_state = DC_IDLE;
#else
    dev = makedev(CDEV_MAJOR,0);
    cdevsw_add(&dev,&mdsio_cdevsw,NULL);

#ifdef DEVFS
    md->devfs_token = devfd_add_devsw(&mdsio_cdevsw,unit,
				      DV_CHR, UID_ROOT, GID_WHEEL, 
				      0660, "mdsio%d", unit);
#endif	/* DEVFS */
#endif	/* FreeBSD_21 */

    return(1);
}

/*
** mdsioopen
**
** Default the UART configuration and enable receiver interrupts.
** Only one opener is allowed at a time. 
*/

int
mdsioopen(dev_t dev, int flag __unused, int mode __unused, struct proc *p __unused)
{
    MDIF	*md;
    int		unit;
    int		s;
    u_short	iobase;
    
    unit = minor(dev);
    if (unit > NMDSIO)			/* not configured */
	return(ENXIO);
    md = mdsio_unit(unit);
    if (md->state & MD_OPEN)		/* already open */
	return(EBUSY);

    s = spltty();
    iobase = md->iobase;
    if (md->hasfifo) {
	/*
	 * (Re)enable and drain fifos.
	 *
	 * Certain SMC chips cause problems if the fifos are enabled
	 * while input is ready.  Turn off the fifo if necessary to
	 * clear the input.  We test the input ready bit after
	 * enabling the fifos to handle races between enabling and
	 * fresh input.  
	 */
	while (TRUE) {
	    outb(iobase + com_fifo, FIFO_RCV_RST | FIFO_XMT_RST | FIFO_ENABLE | MDSIO_FTL);
	    DELAY(100);
	    if (!(inb(iobase + com_lsr) & LSR_RXRDY))
		break;
	    outb(iobase + com_fifo, 0);
	    DELAY(100);
	    (void) inb(iobase + com_data);
	}
    }
    md->rxtail = md->rxhead = 0;		/* empty ringbuffers */
    md->txtail = md->txhead = 0;
    md->eflag = 0;				/* no errors */
    md->rsel.si_flags = 0;			/* nothing selecting */
    md->rsel.si_pid = 0;
    md->wsel.si_flags = 0;
    md->wsel.si_pid = 0;
    md->rxbytes = md->txbytes = 0;
    
    md->state = MD_OPEN;			/* port is now open, no other flags set */
    md->speed = 9600;				/* default to 9600 bps */

    disable_intr();				/* let nothing happen here */
    outb(iobase + com_lctl, CFCR_DLAB);
    outb(iobase + com_dlbl, COMBRD(9600) & 0xff);
    outb(iobase + com_dlbh, (u_int) COMBRD(9600) >> 8);
    outb(iobase + com_lctl, LCR_NORM);		/* = 8 bits, parity enabled, space parity */
    set_mcr(md,MD_RX);				/* set for receive */
    (void) inb(iobase + com_lsr);
    (void) inb(iobase + com_data);
    outb(iobase + com_ier, IER_ERXRDY | IER_ETXRDY);	/* enable interrupts */
    enable_intr();
    timeout(mdsiocomplain, md, hz/2);		/* run the whiner timeout */

    splx(s);
#ifdef FreeBSD_21
    kdc_mdsio[unit].kdc_state = DC_BUSY;
#endif
    debug("OK");
    
    return(0);
}

/*
** mdsioclose
**
** Disable interrupts and mark the port as closed.
*/

int
mdsioclose(dev_t dev, int flag __unused, int mode __unused, struct proc *p __unused)
{
    MDIF	*md;
    int		unit;
    int		s;
    int		rv;
    u_short	iobase;
    
    unit = minor(dev);
    if (unit > NMDSIO)			/* not configured */
	return(ENXIO);
    md = mdsio_unit(unit);
    if (!(md->state & MD_OPEN))		/* XXX not open; what to return? */
	return(ENXIO);
    iobase = md->iobase;

    /* XXX should sleep here to wait for TX output to finish?? */
    s = spltty();

    while (txbuflen(md)) {		/* wait for buffer to empty */
	debug("sleep waiting for tx drain (%d)",txbuflen(md));
	md->state |= MD_DRAINSLEEP;	/* waiting for write activity */
    	rv = tsleep((caddr_t)md, PRIBIO | PCATCH, "mdsiowrite", 0);
	md->state &= ~MD_DRAINSLEEP;	/* yawn */
	if (rv == -1)			/* -1 means interrupted*/
	    break;
    }

    outb(iobase + com_ier, 0);		/* disable interrupts */
    set_mcr(md,MD_RX);			/* disable drive */
    md->state = 0;			/* port closed */
#ifdef FreeBSD_21
    kdc_mdsio[unit].kdc_state = DC_IDLE;
#endif
    untimeout(mdsiocomplain, md);	/* kill the whiner timeout */
    splx(s);
    debug("OK");

    return(0);
}


/*
** mdsioread
**
** Suck data out of the ringbuffer and give it to the caller.
**
** We could be bastards and only copy the next contiguous fragment,
** but in the long run that's probably about as inefficient as
** using a bouncebuffer, so we do.
*/

int
mdsioread(dev_t dev, struct uio *uio, int flag)
{
    MDIF	*md;
    int		unit;
    int		s;
    int		rv = 0;
    u_int	hmuch;
    int		frag;
   
    unit = minor(dev);
    if (unit > NMDSIO)			/* not configured */
	return(ENXIO);
    md = mdsio_unit(unit);

    if (uio->uio_resid % sizeof(u_short))	/* reads must be a multiple of short size */
	return(EINVAL);

    s = spltty();
    if (rxbuflen(md) == 0) {		/* no data ready, must sleep */
	if (flag & IO_NDELAY) {		/* can't sleep, return to caller */
	    splx(s);
	    return(EAGAIN);
	}
	md->state |= MD_READSLEEP;	/* make it so */
	rv = tsleep((caddr_t)md, PRIBIO | PCATCH, "mdsioread", 0);
	md->state &= ~MD_READSLEEP;	/* yawn */
    }
    splx(s);
    if (rv)				/* nonzero means interrupted? */
	return(EINTR);

    /*
    ** There's some data to be read; hand it back.
    */

    hmuch = rxbuflen(md);					/* this much is available (shorts) */
    if ((uio->uio_resid / sizeof(u_short)) < hmuch)		/* compare with requested read size */
	hmuch = (uio->uio_resid / sizeof(u_short));		/* only want this much */
    
    if ((md->rxtail + hmuch) > MDSIO_RXBUF) {			/* buffer wrap */
	frag = (md->rxtail + hmuch) - MDSIO_RXBUF;		/* size of fragment before split (shorts) */
	rv = uiomove((caddr_t)(md->rxbuf + md->rxtail), 	/* move first fragment */
		     frag * sizeof(u_short), uio);
	if (!rv) 						/* OK so far? */
	    rv = uiomove((caddr_t)md->rxbuf,			/* move second fragment */
			 (hmuch - frag) * sizeof(u_short), uio);
    } else {
	rv = uiomove((caddr_t)(md->rxbuf + md->rxtail),		/* move all together */
		     hmuch * sizeof(u_short), uio);
    }
    if (rv) {
	debug("uiomove returned %d",rv);
    } else {
	md->rxtail = (md->rxtail + hmuch) % MDSIO_RXBUF;	/* read OK, move the tail */
	debug("OK");
    }
    return(rv);
}

/*
** mdsiowrite
**
** Stuff outgoing data into the transmit ringbuffer, kick things off
** appropriately to make sure it goes.
*/

int
mdsiowrite(dev_t dev, struct uio *uio, int flag)
{
    MDIF	*md;
    int		unit;
    int		s;
    int		rv;
    int		hmuch,frag;
 
    unit = minor(dev);
    if (unit > NMDSIO)				/* not configured */
	return(ENXIO);
    md = mdsio_unit(unit);

    if ((uio->uio_resid % sizeof(u_short)) ||			/* writes must be multiple of u_short size */
	(uio->uio_resid > (int)(MDSIO_TXBUF*sizeof(u_short)))) {/* and must be smaller than the buffer size */
	debug("attempt to write %d bytes rejected",uio->uio_resid);
	return(EINVAL);
    }
    hmuch = uio->uio_resid / sizeof(short);	/* size of write operation (shorts) */
    s = spltty();				/* tail/head updates maybe not atomic? */
    while (hmuch > txbuffree(md)) {		/* wait for enough buffer space */
	if (flag & IO_NDELAY) {			/* can't sleep */
	    splx(s);
	    return(EAGAIN);
	}
	debug("sleep waiting for write space (have %d, want %d)",txbuffree(md),hmuch);
	md->state |= MD_WRITESLEEP;		/* waiting for a write to make room */
    	rv = tsleep((caddr_t)md, PRIBIO | PCATCH, "mdsiowrite", 0);
	md->state &= ~MD_WRITESLEEP;		/* yawn */
	if (rv)	{				/* nonzero means interrupted? */
	    splx(s);
	    return(EINTR);
	}
    }
    splx(s);
    
    if ((md->txhead + hmuch) > MDSIO_TXBUF) {			/* buffer wrap? */
	frag = MDSIO_TXBUF - md->txhead;			/* space at end */
	rv = uiomove((caddr_t)(md->txbuf + md->txhead),		/* move first fragment */
		     frag * sizeof(u_short), uio);
	if (!rv)						/* all OK so far */
	    rv = uiomove((caddr_t)md->txbuf,			/* move second fragment */
			 (hmuch * frag) * sizeof(u_short), uio);
    } else {
	rv = uiomove((caddr_t)(md->txbuf+md->txhead),hmuch * sizeof(u_short),uio);
    }
    if (rv) {
	debug("uiomove returned %d",rv);
	return(rv);
    }
    md->txhead = (md->txhead + hmuch) % MDSIO_TXBUF;	/* update head pointer */

    disable_intr();
    set_mcr(md,MD_TX);						/* go to Tx mode regardless */
    mdsiointr(unit);						/* simulate interrupt to start xmit */
    enable_intr();

    md->state |= MD_XMITTING;					/* set before checking for timeout */
    md->state &= ~MD_TAILDRAIN;					/* clear taildrain flag */
    if (!(md->state & MD_TIMER)) {				/* no timeout looking after this port? */
	timeout(mdsiotimeout, (void *)md, hz/MDSIO_TIMEOUT);	/* register timeout for this port */
	md->state |= MD_TIMER;
    }
    debug("OK");

    return(0);							/* our work here is done */
}


/*
** mdsiointr
**
** Interrupt handler
**
** Data received while transmitting is discarded, data received with 
** parity errors (address markers) has bit 8 set in the buffer.  Data
** received with framing errors are discarded (can be caused by another
** device coming onto/going off the bus, or noise).
**
** XXX for RS-422 use, the rx/tx interlock should be removed.
*/
    
void
mdsiointr(int unit)
{
    MDIF	*md;
    u_char	line_status;
    u_short	iobase;
    u_short	rdata;
    int		readsome = 0, wrotesome = 0;

    md = mdsio_unit(unit);
    if (!(md->state & MD_OPEN))				/* shouldn't interrupt while closed! */
    {
	error("interrupt while closed");
	return;
    }
    iobase = md->iobase;
    md->ints++;
    

    for (;;) {						/* loop until no interrupting conditions */
	line_status = inb(iobase + com_lsr);

	while(line_status & LSR_RXRDY) {		/* loop reading first */
	    if (line_status & LSR_OE) {			/* overrun? */
		md->eflag |= MD_FIFOOF;
		md->errors ++;
	    }
	    if (line_status & LSR_FE) {			/* framing error? */
		md->eflag |= MD_FERROR;			/* probably worth mentioning */
		md->errors++;
	    }
	    if (line_status & LSR_RXRDY) {
		rdata = inb(iobase + com_data);
		if (line_status & LSR_PE)			/* address marker */
		    rdata |= MD_ADDRMARK;
		if (rxheadnext(md) == md->rxtail) {		/* buffer full 8( */
		    md->eflag |= MD_IBUFOF;			/* mark for later whining */
		    md->errors++;
		} else {
			if ((md->state & MD_XMITTING) &&	/* transmitting */ 
			    (md->options & MD_2WIRE)) {		/* in 2-wire mode */
			    md->eflag |= MD_RXWHILETX;		/* shouldn't happen ?? */
			    md->errors++;
			} else {
			    md->rxbuf[md->rxhead] = rdata;	/* save in buffer */
			    md->rxhead = rxheadnext(md);
			    readsome++;				/* flag for later wakeup of reader */
			}
		}
	    }
	    line_status = inb(iobase + com_lsr);
	    md->rxbytes += readsome;
	}						/* end read loop */
	if (readsome) {					/* anything in? */
	    if (md->rsel.si_pid) {
		selwakeup(&md->rsel);
	    }
	    if (md->state & MD_READSLEEP) {		/* sleeping on read? */
		md->state &= ~MD_READSLEEP;		/* not anymore */
		wakeup((caddr_t)md);			/* kick */
	    }
	}

	if ((line_status & LSR_TXRDY) && 		/* UART ready for tx */
	    !(md->state & (MD_PREADDR|MD_POSTADDR)) && 	/* not held up on an address marker */
	    (md->state & MD_XMITTING)) {		/* we are actually doing something */

	    if (!txbuflen(md)) {			/* nothing to send */
		md->state |= MD_TAILDRAIN;		/* tell timeout to disable Tx later */
		goto tx_done;
	    }
	    if (md->txbuf[md->txtail] & MD_ADDRMARK) {	/* next byte is an address marker? */
		md->state |= MD_PREADDR;		/* tell the timeout code */
		goto tx_done;
	    }
	    outb(iobase + com_data,md->txbuf[md->txtail]&0xff);	/* write the low byte (data) */
	    md->txtail = txtailnext(md);
	    wrotesome++;				/* flag for later wakeup of writer */
	    md->txbytes += wrotesome;
	}
    tx_done:
	if (wrotesome) {				/* anything out? */
	    if (md->wsel.si_pid)
		selwakeup(&md->wsel);
	    if (md->state & MD_WRITESLEEP) {		/* sleeping on write? */
		md->state &= ~MD_WRITESLEEP;		/* not anymore */
		wakeup((caddr_t)md);			/* kick */
	    }
	}
	if ((inb(iobase + com_iir) & 0xf) == 1)		/* no activity pending */
	    return;
    }
}

/*
** mdsiotimeout
**
** Runs while transmitting, and handles waiting for the TX shift register
** to empty before and after address markers. 
**
** XXX may reqire spltty() protection?
*/

static void
mdsiotimeout(void *what)
{
    MDIF	*md;
    u_short	iobase;
    u_char	line_status;

    md = (MDIF *)what;				/* this is what we work with */
    if (md->state & MD_XMITTING) {		/* timeout still wanted? */
	timeout(mdsiotimeout, (void *)md, hz/MDSIO_TIMEOUT);	/* register for another timeout */
    } else {
	md->state &= ~MD_TIMER;			/* mark the timeout as not running */
	return;					/* timeout off now */
    }

    iobase = md->iobase;
    line_status = inb(iobase + com_lsr);	/* get the line status */
    if (!(line_status & LSR_TSRE))		/* shift register empty? */
	return;					/* nothing to do here, get out fast */
	
    if (md->state & MD_PREADDR) {		/* waiting for drain before address byte */
	outb(iobase + com_lctl, LCR_ADDR);	/* = 8 bits, parity enabled, mark parity */
	outb(iobase + com_data,md->txbuf[md->txtail]&0xff);	/* write the low byte (data) */
	md->txtail = txtailnext(md);		/* indicate that the byte's gone */
	md->state |= MD_POSTADDR;		/* now waiting for it to clear */
	md->state &= ~MD_PREADDR;
	if (md->wsel.si_pid)			/* selector on write? */
	    selwakeup(&md->wsel);
	if (md->state & MD_WRITESLEEP) {	/* sleeping on write? */
	    md->state &= ~MD_WRITESLEEP;	/* not anymore */
	    wakeup((caddr_t)md);		/* kick */
	}
	return;
    }
    if (md->state & MD_POSTADDR) {		/* waiting for address byte to drain */
	outb(iobase + com_lctl, LCR_NORM);	/* = 8 bits, parity enabled, space parity */
	md->state &= ~MD_POSTADDR;		/* not waiting anymore */
	disable_intr();
	mdsiointr(md->unit);			/* fake an interrupt for this unit to get tx going */
	enable_intr();
	return;
    }
    if (md->state & MD_TAILDRAIN) {		/* waiting for last byte to go? */
	set_mcr(md,MD_RX);			/* safe to let the bus go now */
	md->state &= ~(MD_TAILDRAIN|MD_XMITTING);	/* clear both states */
	if (md->state & MD_DRAINSLEEP) {	/* sleeping on drain? */
	    md->state &= ~MD_DRAINSLEEP;	/* not anymore */
	    wakeup((caddr_t)md);		/* kick */
	}
	return;
    }
    
    disable_intr();
    mdsiointr(md->unit);			/* if we get here, a transmit interrupt has been lost */
    enable_intr();
}

/*
** mdsioioctl
**
** Ye Olde Ioctl handler.
**
** Not much you can do with one of these ports apart from set the baudrate.
*/
    
int
mdsioioctl(dev_t dev, int cmd, caddr_t data, int flag __unused, struct proc *p __unused)
{
    MDIF	*md;
    int		unit;
    u_short	iobase;
    int		s;
    int		*val = (int *)data;
    u_char	lcr_image;
		
    unit = minor(dev);
    if (unit > NMDSIO)		/* not configured */
	return(ENXIO);
    md = mdsio_unit(unit);
    iobase = md->iobase;

    switch(cmd) {
    case MDSIOCSETSPEED:
	switch(*val)
	{
	case 1200:		/* reasonable values */
	case 2400:
	case 4800:
	case 9600:
	case 19200:
	case 38400:
	case 57600:
	case 115200:
	    break;
	    
	default:		/* unreasonable values */
	    return(EINVAL);
	}

	s = spltty();		/* would be Very Bad to have an interrupt */
	lcr_image = inb(iobase + com_lctl);
	outb(iobase + com_lctl, CFCR_DLAB);
	outb(iobase + com_dlbl, COMBRD(*val) & 0xff);
	outb(iobase + com_dlbh, (u_int) COMBRD(*val) >> 8);
	outb(iobase + com_lctl, lcr_image);
	splx(s);
	return(0);
	
    case MDSIOCGETSPEED:
	*val = md->speed;
	return(0);

    case FIONBIO:		/* ignore these */
    case FIOASYNC:
	return(0);
	
    }
    error("got unknown cmd 0x%x",cmd);
    return(ENOTTY);
}

/*
** mdsioselect
**
** Determine whether there's data available to read/ room to write
*/

int 
mdsioselect(dev_t dev, int rw, struct proc *p)
{
    MDIF	*md;
    int		unit;
    int		s;
    
    
    unit = minor(dev);
    md = mdsio_unit(unit);

    s = spltty();			/* lock buffer stats */
    switch(rw) {
    case FWRITE:			/* selecting for output */
	if (txbuffree(md)) {		/* room to write now */
	    splx(s);
	    return(1);
	}
	selrecord(p,&(md->wsel));
	splx(s);
	return(0);

    case FREAD:				/* selecting to read */
	if (rxbuflen(md)) {		/* data available to read */
	    splx(s);
	    return(1);
	}
	selrecord(p,&(md->rsel));
	splx(s);
	return(0);

    default:
	error("called with invalid rw value (%d)",rw);
	return(1);
    }
}

/*
** mdsiocomplain
**
** periodic error reporter, runs while a port is open
*/

static void
mdsiocomplain(void *what)
{
    MDIF	*md = what;
    int		i,j,s;
    
    timeout(mdsiocomplain, md, hz/2);	/* every half a second */
    
    s = spltty();
    i = md->eflag;			/* get error flag safely */
    md->eflag = 0;
    j = md->errors;
    splx(s);

    if (i & MD_FIFOOF)
	error("FIFO overflow");
    if (i & MD_IBUFOF)
	error("interrupt-level buffer overflow");
    if (i & MD_FERROR)
	error("framing error");
    if (i & MD_RXWHILETX)
	error("received data while transmitting");
    if (i)
	error("total errors : %d",j);
}

#endif
